@@ -15,9 +15,14 @@ package handler
 
 import (
 	"bytes"
+	"errors"
+	"fmt"
 	"net/http"
 	"net/http/httptest"
 	"testing"
+	"time"
+
+	"github.com/go-kit/kit/log"
 
 	"github.com/golang/protobuf/proto"
 	"github.com/julienschmidt/httprouter"
@@ -27,12 +32,28 @@ import (
 	"github.com/prometheus/pushgateway/storage"
 )
 
+var logger = log.NewNopLogger()
+
+// MockMetricStore isn't doing any of the validation and sanitation a real
+// metric store implementation has to do. Those are tested in the storage
+// package. Here we only ensure that the right method calls are performed
+// by the code in the handlers.
 type MockMetricStore struct {
 	lastWriteRequest storage.WriteRequest
+	metricGroups     storage.GroupingKeyToMetricGroup
+	writeRequests    []storage.WriteRequest
+	err              error // If non-nil, will be sent to Done channel in request.
 }
 
 func (m *MockMetricStore) SubmitWriteRequest(req storage.WriteRequest) {
+	m.writeRequests = append(m.writeRequests, req)
 	m.lastWriteRequest = req
+	if req.Done != nil {
+		if m.err != nil {
+			req.Done <- m.err
+		}
+		close(req.Done)
+	}
 }
 
 func (m *MockMetricStore) GetMetricFamilies() []*dto.MetricFamily {
@@ -40,7 +61,7 @@ func (m *MockMetricStore) GetMetricFamilies() []*dto.MetricFamily {
 }
 
 func (m *MockMetricStore) GetMetricFamiliesMap() storage.GroupingKeyToMetricGroup {
-	panic("not implemented")
+	return m.metricGroups
 }
 
 func (m *MockMetricStore) Shutdown() error {
@@ -79,7 +100,10 @@ func TestHealthyReady(t *testing.T) {
 
 func TestPush(t *testing.T) {
 	mms := MockMetricStore{}
-	handler := Push(&mms, false)
+	mmsWithErr := MockMetricStore{err: errors.New("testerror")}
+	handler := Push(&mms, false, false, logger)
+	handlerWithErr := Push(&mmsWithErr, false, false, logger)
+	handlerBase64 := Push(&mms, false, true, logger)
 	req, err := http.NewRequest("POST", "http://example.org/", &bytes.Buffer{})
 	if err != nil {
 		t.Fatal(err)
@@ -99,7 +123,7 @@ func TestPush(t *testing.T) {
 	mms.lastWriteRequest = storage.WriteRequest{}
 	w = httptest.NewRecorder()
 	handler(w, req, httprouter.Params{httprouter.Param{Key: "job", Value: "testjob"}})
-	if expected, got := http.StatusAccepted, w.Code; expected != got {
+	if expected, got := http.StatusOK, w.Code; expected != got {
 		t.Errorf("Wanted status code %v, got %v.", expected, got)
 	}
 	if mms.lastWriteRequest.Timestamp.IsZero() {
@@ -140,7 +164,7 @@ func TestPush(t *testing.T) {
 	mms.lastWriteRequest = storage.WriteRequest{}
 	req, err = http.NewRequest(
 		"POST", "http://example.org/",
-		bytes.NewBufferString("some_metric 3.14\nanother_metric 42\n"),
+		bytes.NewBufferString("some_metric 3.14\nanother_metric{instance=\"testinstance\",job=\"testjob\"} 42\n"),
 	)
 	if err != nil {
 		t.Fatal(err)
@@ -153,7 +177,7 @@ func TestPush(t *testing.T) {
 			httprouter.Param{Key: "labels", Value: "/instance/testinstance"},
 		},
 	)
-	if expected, got := http.StatusAccepted, w.Code; expected != got {
+	if expected, got := http.StatusOK, w.Code; expected != got {
 		t.Errorf("Wanted status code %v, got %v.", expected, got)
 	}
 	if mms.lastWriteRequest.Timestamp.IsZero() {
@@ -165,21 +189,93 @@ func TestPush(t *testing.T) {
 	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
 		t.Errorf("Wanted instance %v, got %v.", expected, got)
 	}
-	if expected, got := `name:"some_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"some_metric" type:UNTYPED metric:<untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
 	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
-	if _, ok := mms.lastWriteRequest.MetricFamilies["push_time_seconds"]; !ok {
-		t.Errorf("Wanted metric family push_time_seconds missing.")
+
+	// With job name and instance name and text content, storage returns error.
+	req, err = http.NewRequest(
+		"POST", "http://example.org/",
+		bytes.NewBufferString("some_metric 3.14\nanother_metric{instance=\"testinstance\",job=\"testjob\"} 42\n"),
+	)
+	if err != nil {
+		t.Fatal(err)
+	}
+	w = httptest.NewRecorder()
+	handlerWithErr(
+		w, req,
+		httprouter.Params{
+			httprouter.Param{Key: "job", Value: "testjob"},
+			httprouter.Param{Key: "labels", Value: "/instance/testinstance"},
+		},
+	)
+	if expected, got := http.StatusBadRequest, w.Code; expected != got {
+		t.Errorf("Wanted status code %v, got %v.", expected, got)
+	}
+	if mmsWithErr.lastWriteRequest.Timestamp.IsZero() {
+		t.Errorf("Write request timestamp not set: %#v", mmsWithErr.lastWriteRequest)
+	}
+	if expected, got := "testjob", mmsWithErr.lastWriteRequest.Labels["job"]; expected != got {
+		t.Errorf("Wanted job %v, got %v.", expected, got)
+	}
+	if expected, got := "testinstance", mmsWithErr.lastWriteRequest.Labels["instance"]; expected != got {
+		t.Errorf("Wanted instance %v, got %v.", expected, got)
+	}
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"some_metric" type:UNTYPED metric:<untyped:<value:3.14 > > `, mmsWithErr.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
+		t.Errorf("Wanted metric family %v, got %v.", expected, got)
+	}
+	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mmsWithErr.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
+		t.Errorf("Wanted metric family %v, got %v.", expected, got)
+	}
+
+	// With base64-encoded job name and instance name and text content.
+	mms.lastWriteRequest = storage.WriteRequest{}
+	req, err = http.NewRequest(
+		"POST", "http://example.org/",
+		bytes.NewBufferString("some_metric 3.14\nanother_metric{instance=\"testinstance\",job=\"testjob\"} 42\n"),
+	)
+	if err != nil {
+		t.Fatal(err)
+	}
+	w = httptest.NewRecorder()
+	handlerBase64(
+		w, req,
+		httprouter.Params{
+			httprouter.Param{Key: "job", Value: "dGVzdC9qb2I="},                         // job="test/job"
+			httprouter.Param{Key: "labels", Value: "/instance@base64/dGVzdGluc3RhbmNl"}, // instance="testinstance"
+		},
+	)
+	if expected, got := http.StatusOK, w.Code; expected != got {
+		t.Errorf("Wanted status code %v, got %v.", expected, got)
+	}
+	if mms.lastWriteRequest.Timestamp.IsZero() {
+		t.Errorf("Write request timestamp not set: %#v", mms.lastWriteRequest)
+	}
+	if expected, got := "test/job", mms.lastWriteRequest.Labels["job"]; expected != got {
+		t.Errorf("Wanted job %v, got %v.", expected, got)
+	}
+	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
+		t.Errorf("Wanted instance %v, got %v.", expected, got)
+	}
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"some_metric" type:UNTYPED metric:<untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
+		t.Errorf("Wanted metric family %v, got %v.", expected, got)
+	}
+	// Note that sanitation hasn't happened yet, job label as still as in the push, not aligned to grouping labels.
+	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
+		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
 
 	// With job name and no instance name and text content.
 	mms.lastWriteRequest = storage.WriteRequest{}
 	req, err = http.NewRequest(
 		"POST", "http://example.org/",
-		bytes.NewBufferString("some_metric 3.14\nanother_metric 42\n"),
+		bytes.NewBufferString("some_metric 3.14\nanother_metric{instance=\"testinstance\",job=\"testjob\"} 42\n"),
 	)
 	if err != nil {
 		t.Fatal(err)
@@ -191,7 +287,7 @@ func TestPush(t *testing.T) {
 			httprouter.Param{Key: "job", Value: "testjob"},
 		},
 	)
-	if expected, got := http.StatusAccepted, w.Code; expected != got {
+	if expected, got := http.StatusOK, w.Code; expected != got {
 		t.Errorf("Wanted status code %v, got %v.", expected, got)
 	}
 	if mms.lastWriteRequest.Timestamp.IsZero() {
@@ -203,10 +299,11 @@ func TestPush(t *testing.T) {
 	if expected, got := "", mms.lastWriteRequest.Labels["instance"]; expected != got {
 		t.Errorf("Wanted instance %v, got %v.", expected, got)
 	}
-	if expected, got := `name:"some_metric" type:UNTYPED metric:<label:<name:"instance" value:"" > label:<name:"job" value:"testjob" > untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"some_metric" type:UNTYPED metric:<untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
-	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
+	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
 
@@ -227,52 +324,19 @@ func TestPush(t *testing.T) {
 			httprouter.Param{Key: "labels", Value: "/instance/testinstance"},
 		},
 	)
-	if expected, got := http.StatusBadRequest, w.Code; expected != got {
+	// Note that a real storage shourd reject pushes with timestamps. Here
+	// we only make sure it gets through. Rejection is tested in the storage
+	// package.
+	if expected, got := http.StatusOK, w.Code; expected != got {
 		t.Errorf("Wanted status code %v, got %v.", expected, got)
 	}
-	if !mms.lastWriteRequest.Timestamp.IsZero() {
-		t.Errorf("Write request timestamp unexpectedly set: %#v", mms.lastWriteRequest)
+	// Make sure the timestamp from the push didn't make it to the WriteRequest.
+	if time.Now().Sub(mms.lastWriteRequest.Timestamp) > time.Minute {
+		t.Errorf("Write request timestamp set to a too low value: %#v", mms.lastWriteRequest)
 	}
-
-	// With job name and instance name and text content and job and instance labels.
-	mms.lastWriteRequest = storage.WriteRequest{}
-	req, err = http.NewRequest(
-		"POST", "http://example.org",
-		bytes.NewBufferString(`
-some_metric{job="foo",instance="bar"} 3.14
-another_metric{instance="baz"} 42
-`),
-	)
-	if err != nil {
-		t.Fatal(err)
+	if expected, got := int64(1000), mms.lastWriteRequest.MetricFamilies["b"].GetMetric()[0].GetTimestampMs(); expected != got {
+		t.Errorf("Wanted protobuf timestamp %v, got %v.", expected, got)
 	}
-	w = httptest.NewRecorder()
-	handler(
-		w, req,
-		httprouter.Params{
-			httprouter.Param{Key: "job", Value: "testjob"},
-			httprouter.Param{Key: "labels", Value: "/instance/testinstance"},
-		},
-	)
-	if expected, got := http.StatusAccepted, w.Code; expected != got {
-		t.Errorf("Wanted status code %v, got %v.", expected, got)
-	}
-	if mms.lastWriteRequest.Timestamp.IsZero() {
-		t.Errorf("Write request timestamp not set: %#v", mms.lastWriteRequest)
-	}
-	if expected, got := "testjob", mms.lastWriteRequest.Labels["job"]; expected != got {
-		t.Errorf("Wanted job %v, got %v.", expected, got)
-	}
-	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
-		t.Errorf("Wanted instance %v, got %v.", expected, got)
-	}
-	if expected, got := `name:"some_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
-		t.Errorf("Wanted metric family %v, got %v.", expected, got)
-	}
-	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:42 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
-		t.Errorf("Wanted metric family %v, got %v.", expected, got)
-	}
-
 	// With job name and instance name and protobuf content.
 	mms.lastWriteRequest = storage.WriteRequest{}
 	buf := &bytes.Buffer{}
@@ -321,7 +385,7 @@ another_metric{instance="baz"} 42
 			httprouter.Param{Key: "labels", Value: "/instance/testinstance"},
 		},
 	)
-	if expected, got := http.StatusAccepted, w.Code; expected != got {
+	if expected, got := http.StatusOK, w.Code; expected != got {
 		t.Errorf("Wanted status code %v, got %v.", expected, got)
 	}
 	if mms.lastWriteRequest.Timestamp.IsZero() {
@@ -333,17 +397,20 @@ another_metric{instance="baz"} 42
 	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
 		t.Errorf("Wanted instance %v, got %v.", expected, got)
 	}
-	if expected, got := `name:"some_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:1.234 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"some_metric" type:UNTYPED metric:<untyped:<value:1.234 > > `, mms.lastWriteRequest.MetricFamilies["some_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
-	if expected, got := `name:"another_metric" type:UNTYPED metric:<label:<name:"instance" value:"testinstance" > label:<name:"job" value:"testjob" > untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
+	// Note that sanitation hasn't happened yet, grouping labels not in request.
+	if expected, got := `name:"another_metric" type:UNTYPED metric:<untyped:<value:3.14 > > `, mms.lastWriteRequest.MetricFamilies["another_metric"].String(); expected != got {
 		t.Errorf("Wanted metric family %v, got %v.", expected, got)
 	}
 }
 
 func TestDelete(t *testing.T) {
 	mms := MockMetricStore{}
-	handler := Delete(&mms)
+	handler := Delete(&mms, false, logger)
+	handlerBase64 := Delete(&mms, true, logger)
 
 	// No job name.
 	mms.lastWriteRequest = storage.WriteRequest{}
@@ -403,37 +470,144 @@ func TestDelete(t *testing.T) {
 	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
 		t.Errorf("Wanted instance %v, got %v.", expected, got)
 	}
+
+	// With base64-encoded job name and instance name.
+	mms.lastWriteRequest = storage.WriteRequest{}
+	w = httptest.NewRecorder()
+	handlerBase64(
+		w, &http.Request{},
+		httprouter.Params{
+			httprouter.Param{Key: "job", Value: "dGVzdC9qb2I="},                         // job="test/job"
+			httprouter.Param{Key: "labels", Value: "/instance@base64/dGVzdGluc3RhbmNl"}, // instance="testinstance"
+		},
+	)
+	if expected, got := http.StatusAccepted, w.Code; expected != got {
+		t.Errorf("Wanted status code %v, got %v.", expected, got)
+	}
+	if mms.lastWriteRequest.Timestamp.IsZero() {
+		t.Errorf("Write request timestamp not set: %#v", mms.lastWriteRequest)
+	}
+	if expected, got := "test/job", mms.lastWriteRequest.Labels["job"]; expected != got {
+		t.Errorf("Wanted job %v, got %v.", expected, got)
+	}
+	if expected, got := "testinstance", mms.lastWriteRequest.Labels["instance"]; expected != got {
+		t.Errorf("Wanted instance %v, got %v.", expected, got)
+	}
+
 }
 
 func TestSplitLabels(t *testing.T) {
-	properLabels := "/label_name1/label_value1/label_name2/label_value2"
-	expectedParsed := map[string]string{
-		"label_name1": "label_value1",
-		"label_name2": "label_value2",
-	}
-	parsed, err := splitLabels(properLabels)
-	if err != nil {
-		t.Errorf("Got unexpected error: %s.", err)
-	}
-	for k, v := range expectedParsed {
-		got, ok := parsed[k]
-		if !ok {
-			t.Errorf("Expected to find key %s.", k)
-		}
-		if got != v {
-			t.Errorf("Expected %s but got %s.", v, got)
-		}
+	scenarios := map[string]struct {
+		input          string
+		expectError    bool
+		expectedOutput map[string]string
+	}{
+		"regular labels": {
+			input: "/label_name1/label_value1/label_name2/label_value2",
+			expectedOutput: map[string]string{
+				"label_name1": "label_value1",
+				"label_name2": "label_value2",
+			},
+		},
+		"invalid label name": {
+			input:       "/label_name1/label_value1/a=b/label_value2",
+			expectError: true,
+		},
+		"reserved label name": {
+			input:       "/label_name1/label_value1/__label_name2/label_value2",
+			expectError: true,
+		},
+		"unencoded slash in label value": {
+			input:       "/label_name1/label_value1/label_name2/label/value2",
+			expectError: true,
+		},
+		"encoded slash in first label value ": {
+			input: "/label_name1@base64/bGFiZWwvdmFsdWUx/label_name2/label_value2",
+			expectedOutput: map[string]string{
+				"label_name1": "label/value1",
+				"label_name2": "label_value2",
+			},
+		},
+		"encoded slash in last label value": {
+			input: "/label_name1/label_value1/label_name2@base64/bGFiZWwvdmFsdWUy",
+			expectedOutput: map[string]string{
+				"label_name1": "label_value1",
+				"label_name2": "label/value2",
+			},
+		},
+		"encoded slash in last label value with padding": {
+			input: "/label_name1/label_value1/label_name2@base64/bGFiZWwvdmFsdWUy==",
+			expectedOutput: map[string]string{
+				"label_name1": "label_value1",
+				"label_name2": "label/value2",
+			},
+		},
+		"invalid base64 encoding": {
+			input:       "/label_name1@base64/foo.bar/label_name2/label_value2",
+			expectError: true,
+		},
 	}
 
-	improperLabels := "/label_name1/label_value1/a=b/label_value2"
-	_, err = splitLabels(improperLabels)
-	if err == nil {
-		t.Error("Expected splitLabels to return an error when given improper labels.")
+	for name, scenario := range scenarios {
+		t.Run(name, func(t *testing.T) {
+			parsed, err := splitLabels(scenario.input)
+			if err != nil {
+				if scenario.expectError {
+					return // All good.
+				}
+				t.Fatalf("Got unexpected error: %s.", err)
+			}
+			for k, v := range scenario.expectedOutput {
+				got, ok := parsed[k]
+				if !ok {
+					t.Errorf("Expected to find %s=%q.", k, v)
+				}
+				if got != v {
+					t.Errorf("Expected %s=%q but got %s=%q.", k, v, k, got)
+				}
+				delete(parsed, k)
+			}
+			for k, v := range parsed {
+				t.Errorf("Found unexpected label %s=%q.", k, v)
+			}
+		})
 	}
+}
 
-	reservedLabels := "/label_name1/label_value1/__label_name2/label_value2"
-	_, err = splitLabels(reservedLabels)
-	if err == nil {
-		t.Error("Expected splitLabels to return an error when given a reserved label.")
+func TestWipeMetricStore(t *testing.T) {
+	// Create MockMetricStore with a few GroupingKeyToMetricGroup metrics
+	// so they can be returned by GetMetricFamiliesMap() to later send write
+	// requests for each of them.
+	metricCount := 5
+	mgs := storage.GroupingKeyToMetricGroup{}
+	for i := 0; i < metricCount; i++ {
+		mgs[fmt.Sprint(i)] = storage.MetricGroup{}
+	}
+	mms := MockMetricStore{metricGroups: mgs}
+
+	// Wipe handler should return 202 and delete all metrics.
+	wipeHandler := WipeMetricStore(&mms, logger)
+	w := httptest.NewRecorder()
+	// Then handler is routed to the handler based on verb and path in main.go
+	// therefore (and for now) we use the request to only record the returned status code.
+	req, err := http.NewRequest("PUT", "http://example.org", &bytes.Buffer{})
+	if err != nil {
+		t.Fatal(err)
+	}
+	wipeHandler.ServeHTTP(w, req)
+
+	if w.Code != http.StatusAccepted {
+		t.Errorf("status code should be %d", http.StatusAccepted)
+	}
+
+	if len(mms.writeRequests) != metricCount {
+		t.Errorf("there should be %d write requests, got %d instead", metricCount, len(mms.writeRequests))
+	}
+
+	// Were all the writeRequest deletes?.
+	for i, wr := range mms.writeRequests {
+		if wr.MetricFamilies != nil {
+			t.Errorf("writeRequest at index %d was not a delete request", i)
+		}
 	}
 }
