@@ -19,6 +19,7 @@ import (
 
 	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/client_golang/prometheus/promhttp"
+	"github.com/prometheus/common/server"
 
 	"github.com/prometheus/pushgateway/storage"
 )
@@ -63,9 +64,17 @@ func Ready(ms storage.MetricStore) http.Handler {
 // Static serves the static files from the provided http.FileSystem.
 //
 // The returned handler is already instrumented for Prometheus.
-func Static(root http.FileSystem) http.Handler {
+func Static(root http.FileSystem, prefix string) http.Handler {
+	if prefix == "/" {
+		prefix = ""
+	}
+
+	handler := server.StaticFileServer(root)
 	return promhttp.InstrumentHandlerCounter(
 		httpCnt.MustCurryWith(prometheus.Labels{"handler": "static"}),
-		http.FileServer(root),
+		http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+			r.URL.Path = r.URL.Path[len(prefix):]
+			handler.ServeHTTP(w, r)
+		}),
 	)
 }
